use std::any::Any;

pub struct Orchestrator {}

impl Orchestrator {
    pub fn cron(&self, cron: impl AsRef<str>) -> Scheduler {
        Scheduler {}
    }

    pub fn query(&self, query: Query) -> Response {
        Response::Tags { list: Vec::new() }
    }

    pub fn api<A: Any>(&self, request: A) -> Response {
        Response::Tags { list: Vec::new() }
    }
}

pub enum Query {
    Tags(String),
}

pub enum Response {
    Tags { list: Vec<String> },
}

#[derive(Clone)]
pub struct Scheduler {}

pub trait Ensure {
    type EnsureInput;
    type EnsureOutput;

    fn ensure(&self, input: Self::EnsureInput) -> Self::EnsureOutput;
}

pub trait Action {
    type Input;
    type Output;

    fn call(&self, input: Self::Input) -> Self::Output;
}

impl Scheduler {
    pub fn schedule<A: Action>(&self, action: A) -> Scheduler {
        return self.clone();
    }

    pub async fn apply(&self) {}
}

pub struct DebianPkg {}

impl DebianPkg {
    pub fn require<'a>(pkgs: impl Into<Vec<&'a str>>) -> Self {
        Self {}
    }
}
impl Ensure for DebianPkg {
    type EnsureInput = ();
    type EnsureOutput = ();

    fn ensure(&self, input: Self::EnsureInput) -> Self::EnsureOutput {}
}

impl Action for DebianPkg {
    type Input = ();
    type Output = ();

    fn call(&self, input: Self::Input) -> Self::Output {}
}

pub async fn handle(orchestrator: Orchestrator) {
    orchestrator
        .cron("* * * * *")
        .schedule(DebianPkg::require(vec!["openssl", "git"]))
        .apply()
        .await;

    orchestrator
        .cron("* * * * *")
        .schedule(DebianPkg::require(vec!["openssl", "git"]))
        .apply()
        .await;
}
